/* $Id: jigdo-file.cc,v 1.34 2001/09/23 22:13:45 richard Exp $ -*- C++ -*-
  __   _
  |_) /|  Copyright (C) 2000 Richard Atterer
  | \/|  <richard@atterer.net>
   ` 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2. See
  the file COPYING for details.

  Command line utility to create .jigdo and .template files

*/

#include <config.h>

#include <memory>
#include <fstream>
#include <sstream>
using namespace std;
#ifdef HAVE_GETOPT_H
#  include <getopt.h>
#else
#  include <glibc-getopt.h>
#endif
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <zlib.h>

#include <compat.hh>
#include <configfile.hh>
#include <debug.hh>
#include <jigdoconfig.hh>
#include <mimestream.hh>
#include <mkimage.hh>
#include <mktemplate.hh>
#include <recursedir.hh>
#include <scan.hh>
#include <string.hh>
//______________________________________________________________________

namespace {

#if WINDOWS
const char* const binaryName = "jigdo-file";
#else
string binaryName; // of the program
#endif

char optHelp = '\0';
bool optVersion = false;
// Names of files given on command line, and of --files-from files
RecurseDir fileNames;
string imageFile;
string jigdoFile;
string templFile;
string cacheFile;
size_t optCacheExpiry = 60*60*24*30; // Expiry time for cache in seconds
vector<string> optLabels; // Strings of the form "Label=/some/path"
vector<string> optUris;   // Strings of the form "Label=http://some.server/"
size_t blockLength     =  64*1024U; // of rsync algorithm
size_t miniBlockLength =   4*1024U; // is also minimum file size
size_t md5BlockLength  = 128*1024U - 55; // a bit more efficient with "- 55"
size_t readAmount      = 128*1024U;
int optZipQuality = Z_BEST_COMPRESSION;
bool optForce = false; // true => Silently delete existent output files
bool optMkImageCheck = true; // true => check MD5sums of files written to img
bool optAddImage = true; // true => Add [Image] section to output .jigdo
bool optAddServers = true; // true => Add [Servers] section to output .jigdo

// Return value of main(), for "delayed error exit"
int returnValue = 0;
// Size of image file (zero if stdin), for percentage progress reports
uint64 imageSize = 0;
//______________________________________________________________________

/// class for "pointer to any *Reporter class", with disambiguation members
struct AnyReporter : public MkTemplate::ProgressReporter,
                     public JigdoCache::ProgressReporter,
                     public JigdoDesc::ProgressReporter,
                     public MD5Sum::ProgressReporter,
                     public JigdoConfig::ProgressReporter {
  virtual void error(const string& message) {
    MD5Sum::ProgressReporter::error(message);
  }
  virtual void info(const string& message) {
    MD5Sum::ProgressReporter::info(message);
  }
};
//____________________

/// Progress report class that writes informational messages to cerr
class MyProgressReporter : public AnyReporter {
public:
  MyProgressReporter(bool prog) : printProgress(prog) { }

  virtual void error(const string& message) {
    //string m(binaryName); m += ": "; m += message;
    //print(m);
    print(message);
  }
  virtual void info(const string& message) {
    //string m(binaryName); m += ": "; m += message;
    //print(m);
    print(message);
  }
  virtual void scanningFile(const FilePart* file, uint64 offInFile) {
    if (!printProgress) return;
    string m;
    append(m, 100 * offInFile / file->size(), 3); // 3
    m += '%'; // 1 char
    append(m, offInFile / 1024, 8); // >= 8 chars
    m += "k/"; // 2 chars
    append(m, file->size() / 1024); // want >= 8 chars
    m += 'k'; // 1 char
    if (m.size() < 3+1+8+2+8+1)
      m += "          " + 10 - (3+1+8+2+8+1 - m.size());
    m += "scanning `";
    m += file->leafName();
    m += '\'';
    print(m, false);
  }
  virtual void scanningImage(uint64 offset) {
    if (!printProgress) return;
    string m;
    if (imageSize != 0) {
      append(m, 100 * (totalAll + offset) / (totalAll + imageSize), 3); // 3
      m += '%'; // 1 char
    } else {
      m += "    ";
    }
    append(m, offset / 1024, 8); // >= 8 chars
    m += 'k'; // 1 char
    if (imageSize != 0) {
      m += '/'; // 1 char
      append(m, imageSize / 1024); // want >= 8 chars
      m += 'k'; // 1 char
    }
    if (m.size() < 3+1+8+1+1+8+1)
      m += "          " + 10 - (3+1+8+1+1+8+1 - m.size());
    m += "scanning image";
    print(m, false);
  }
  virtual void readingMD5(uint64 offInStream, uint64 size) {
    if (!printProgress) return;
    string m;
    append(m, 100 * offInStream / size, 3); // 3
    m += '%'; // 1 char
    append(m, offInStream / 1024, 8); // >= 8 chars
    m += "k/"; // 2 chars
    append(m, size / 1024); // want >= 8 chars
    m += 'k'; // 1 char
    if (m.size() < 3+1+8+2+8+1)
      m += "          " + 10 - (3+1+8+2+8+1 - m.size());
    m += "verifying image";
    print(m, false);
  }
  virtual void writingImage(uint64 written, uint64 totalToWrite,
                            uint64 imgOff, uint64 imgSize) {
    if (!printProgress) return;
    string m;
    append(m, 100 * written / totalToWrite, 3); // 3
    m += '%'; // 1 char
    append(m, imgOff / 1024, 8); // >= 8 chars
    m += "k/"; // 2 chars
    append(m, imgSize / 1024); // want >= 8 chars
    m += 'k'; // 1 char
    if (m.size() < 3+1+8+2+8+1)
      m += "          " + 10 - (3+1+8+2+8+1 - m.size());
    m += "writing image";
    print(m, false);
  }
  virtual void abortingScan() {
    string m("Error scanning image - abort");
    print(m);
  }
  virtual void matchFound(const FilePart* file, uint64 offInImage) {
    string m("Match of `");
    m += file->leafName();
    m += "' at offset ";
    append(m, offInImage);
    print(m);
  }
  virtual void finished(uint64 imageSize) {
    if (!printProgress) return;
    string m("Finished - image size is ");
    append(m, imageSize);
    m += " bytes.";
    print(m);
  }
private:
  static void print(const string& s, bool addNewline = true);
  bool printProgress;
  static string prevLine;
  static uint64 totalAll;
};
//________________________________________

string MyProgressReporter::prevLine;
uint64 MyProgressReporter::totalAll = 0;

/* Print the string to stderr. Repeatedly overwrite the same line with
   different progress reports if requested with addNewline==false. If
   the tail of the message is the same as a previous message printed
   on the same line, do not print the tail. */
void MyProgressReporter::print(const string& s, bool addNewline) {
  if (s.size() != prevLine.size()) {
    // Print new message, maybe pad with spaces to overwrite rest of old one
    cerr << s;
    if (s.size() < prevLine.size()) {
      size_t nrSpaces = prevLine.size() - s.size();
      while (nrSpaces >= 10) { cerr << "          "; nrSpaces -= 10; }
      if (nrSpaces > 0) cerr << ("         " + 9 - nrSpaces);
    }
  } else {
    /* Same length as previous - to reduce cursor flicker, only redraw
       as much as necessary */
    int i = s.size() - 1;
    while (i >= 0 && s[i] == prevLine[i]) --i;
    for (int j = 0; j <= i; ++j) cerr << s[j];
  }

  // Should the message just printed be overwritten on next call?
  if (addNewline) {
    // No, leave it visible
    cerr << endl;
    prevLine = "";
  } else {
    // Yes, overwrite with next message
    cerr << '\r';
    prevLine = s;
  }
}
//________________________________________

/// Progress report class that writes informational messages to cerr
class MyGrepProgressReporter : public AnyReporter {
public:
  // Default error()/info() will print to cerr
  virtual void matchFound(const FilePart* file, uint64 offInImage) {
    cout << offInImage << ' ' << file->getPath() << file->leafName() << endl;
  }
};
//________________________________________

/// Progress report class that writes informational messages to cerr
class MyQuietProgressReporter : public AnyReporter {
  virtual void info(const string&) { }
};
//________________________________________

MyProgressReporter reporterDefault(true);
MyProgressReporter reporterNoprogress(false);
MyGrepProgressReporter reporterGrep;
MyQuietProgressReporter reporterQuiet;

// Reporter is defined in config.h and is the base of all other *Reporter's
AnyReporter* optReporter = &reporterDefault;
//______________________________________________________________________

inline void printUsage(bool detailed) {
  if (detailed) {
    cout <<
    "\n"
    "Copyright (C) 2001 Richard Atterer <richard@atterer.net>\n"
    "This program is free software; you can redistribute it and/or modify\n"
    "it under the terms of the GNU General Public License, version 2. See\n"
    "the file COPYING or <http://www.gnu.org/copyleft/gpl.html> for details.\n"
    "\n";
  }
  cout <<
    "\n"
    "Usage: " << binaryName << " COMMAND [OPTIONS] [FILES...]\n"
    "Commands:\n"
    "  make-template mt Create template and jigdo from image and files\n"
    "  make-image mi    Recreate image from template and files (can merge\n"
    "                   files in >1 steps, uses `IMG"EXTSEPS"tmp' for --image=IMG)\n"
    ;
  if (detailed) cout <<
    "  print-missing pm After make-image, print files still missing for\n"
    "                   the image to be completely recreated\n";
  cout <<
    "  verify ver       Check whether image matches checksum from template\n"
    "  scan sc          Update cache with information about supplied files\n"
    ;
  if (detailed) {
    cout <<
    "  list-template ls Print low-level listing of contents of template\n"
    "                   data or "EXTSEPS"tmp file\n";
  }
  cout <<
    "\n"
    "Important options:\n"
    "  -i  --image=FILE Output/input filename for image file\n"
    "  -j  --jigdo=FILE Input/output filename for jigdo file\n"
    "  -t  --template=FILE\n"
    "                   Input/output filename for template file\n"
    "  -T  --files-from=FILE\n"
    "                   Read further filenames from FILE (`-' for stdin)\n"
    "  -r  --report=default|noprogress|quiet|grep\n"
    "                   Control format of status reports to stderr (or\n"
    "                   stdout in case of `grep')\n"
    "  -f  --force      Silently delete existent output files\n"
#if WINDOWS
    "      --label Label=C:\\path\n"
    "                   [make-template] Replace name of input file\n"
    "                   `C:\\path\\.\\a\\file.txt' (note the `\\.\\') with\n"
    "                   `Label:a/file.txt' in output jigdo\n"
#else
    "      --label Label=/path\n"
    "                   [make-template] Replace name of input file\n"
    "                   `/path//a/file"EXTSEPS"txt' (note the `//') with\n"
    "                   `Label:a/file"EXTSEPS"txt' in output jigdo\n"
#endif
    "      --uri Label=http://www.site.com\n"
    "                   [make-template] Add mapping from Label to given\n"
    "                   URI instead of default `file:' URI.\n"
    "                   [print-missing] Override mapping in input jigdo\n"
    "  -0 to -9         Set amount of compression in output template\n"
    "      --cache=FILE Store/reload information about any files scanned\n";
  if (detailed) {
    cout <<
      "      --no-cache   Do not cache information about scanned files\n"
      "      --cache-expiry=SECONDS[h|d|w|m|y]\n"
      "                   Remove cache entries if last access was longer\n"
      "                   ago than given amount of time [default 30 days]\n"
      "  -h  --help       Output short help.\n"
      "  -H  --help-all   Output this help.\n";
  } else {
    cout << "  -h  --help       Output this help.\n"
            "  -H  --help-all   Output more detailed help.\n";
  }
  cout << "  -v  --version    Output version info." << endl;
  if (detailed) {
    cout <<
    "\n"
    "Further options: (can append 'k', 'M', 'G' to any BYTES argument)\n"
    "  --no-force   Do not delete existent output files [default]\n"
    "  --min-length=BYTES [default " << miniBlockLength << "]\n"
    "               [make-template] Minimum length of files to search for\n"
    "               in image data\n"
    "  --block-size=BYTES [default " << blockLength << "]\n"
    "               [make-template] Uninteresting internal parameter\n"
    "  --md5-block-size=BYTES [default " << md5BlockLength << "]\n"
    "               Uninteresting internal parameter\n"
    "  jigdo-file enforces: min-length <= block-size < md5-block-size\n"
    "  --readbuffer=BYTES [default " << readAmount / 1024 << "k]\n"
    "               Amount of data to read at a time\n"
    "  --check-files [default]\n"
    "               [make-image] Verify checksum of files written to image\n"
    "  --no-check-files\n"
    "               [make-image] Do not verify checksums\n"
    "  --image-section [default]\n"
    "  --no-image-section\n"
    "  --servers-section [default]\n"
    "  --no-servers-section\n"
    "               [make-template] When creating the jigdo file, do or\n"
    "               do not add the sections \"[Image]\" or \"[Servers]\"."
         << endl;
  }
  return;
}
//______________________________________________________________________

void exit_tryHelp() {
  cerr << binaryName << ": Try `" << binaryName
       << " --help' or `man jigdo-file' for more information" << endl;
  throw Cleanup(3);
}
//______________________________________________________________________

/* Like atoi(), but tolerate exactly one suffix 'k', 'K', 'm', 'M',
   'g' or 'G' for kilo, mega, giga */
size_t scanMemSize(const char* str) {
  const char* s = str;
  size_t x = 0;
  while (*s >= '0' && *s <= '9') {
    x = 10 * x + *s - '0';
    ++s;
  }
  switch (*s) { // Fallthrough mania!
  case 'g': case 'G':
    x = x * 1024;
  case 'm': case 'M':
    x = x * 1024;
  case 'k': case 'K':
    x = x * 1024;
    if (*++s == '\0') return x;
  default:
    cerr << binaryName << ": Invalid size specifier `" << str << '\''
         << endl;
    throw Cleanup(3);
  case '\0': return x;
  }
}

/* Like atoi(), but tolerate exactly one suffix 'h', 'd', 'w', 'm',
   'y' for hours, days, weeks, months, years. Returns seconds. Special
   value "off" results in 0 to be returned. */
size_t scanTimespan(const char* str) {
  if (strcmp(str, "off") == 0) return 0;
  const char* s = str;
  size_t x = 0;
  while (*s >= '0' && *s <= '9') {
    x = 10 * x + *s - '0';
    ++s;
  }
  switch (*s) {
  case 'h': case 'H': x = x * 60 * 60; ++s; break;
  case 'd': case 'D': x = x * 60 * 60 * 24; ++s; break;
  case 'w': case 'W': x = x * 60 * 60 * 24 * 7; ++s; break;
  case 'm': case 'M': x = x * 60 * 60 * 24 * 30; ++s; break;
  case 'y': case 'Y': x = x * 60 * 60 * 24 * 365; ++s; break;
  }
  if (*s == '\0') return x;
  cerr << binaryName << ": Invalid time specifier `" << str << '\''
       << endl;
  throw Cleanup(3);
}
//______________________________________________________________________

/* Try creating a filename in dest by stripping any file extension
   from try1 and appending ext. try1 mustn't be empty or "-", if it
   is, fall back to try2. If try2 is also empty or "-", do not modify
   dest. Sets dest to "" if it would otherwise be equal to either of
   try1, try2. */
void deduceName(string& dest, const char* ext,
                const string& try1, const string& try2) {
  const string* src;
  if (!try1.empty() && try1 != "-") src = &try1;
  else if (!try2.empty() && try2 != "-") src = &try2;
  else return;

  string::size_type lastDot = src->find_last_of(EXTSEP);
  if (src->find_first_of(DIRSEP, lastDot + 1) != string::npos)
    lastDot = string::npos;
  dest.assign(*src, 0U, lastDot);
  dest += ext;
  if (dest == try1 || dest == try2) dest = "";
}
//______________________________________________________________________

enum {
  LONGOPT_BUFSIZE = 0x100, LONGOPT_NOFORCE, LONGOPT_MINSIZE,
  LONGOPT_BLOCKSIZE, LONGOPT_MD5SIZE, LONGOPT_MKIMAGECHECK,
  LONGOPT_NOMKIMAGECHECK, LONGOPT_LABEL, LONGOPT_URI, LONGOPT_ADDSERVERS,
  LONGOPT_NOADDSERVERS, LONGOPT_ADDIMAGE, LONGOPT_NOADDIMAGE,
  LONGOPT_NOCACHE, LONGOPT_CACHEEXPIRY
};

enum Command {
  MAKE_TEMPLATE, MAKE_IMAGE, PRINT_MISSING, SCAN, VERIFY, LIST_TEMPLATE
};

/// Deal with command line switches
Command cmdOptions(int argc, char* argv[]) {
# if !WINDOWS
  binaryName = argv[0];
# endif
  bool error = false;

  while (true) {
    static const struct option longopts[] = {
      { "help",               no_argument,       0, 'h' },
      { "help-all",           no_argument,       0, 'H' },
      { "version",            no_argument,       0, 'v' },
      { "files-from",         required_argument, 0, 'T' }, // "-T" like tar's
      { "image",              required_argument, 0, 'i' },
      { "jigdo",              required_argument, 0, 'j' },
      { "template",           required_argument, 0, 't' },
      { "cache",              required_argument, 0, 'c' },
      { "cache-expiry",       required_argument, 0, LONGOPT_CACHEEXPIRY },
      { "no-cache",           no_argument,       0, LONGOPT_NOCACHE },
      { "force",              no_argument,       0, 'f' },
      { "no-force",           no_argument,       0, LONGOPT_NOFORCE },
      { "min-length",         required_argument, 0, LONGOPT_MINSIZE },
      { "block-size",         required_argument, 0, LONGOPT_BLOCKSIZE },
      { "md5-block-size",     required_argument, 0, LONGOPT_MD5SIZE },
      { "readbuffer",         required_argument, 0, LONGOPT_BUFSIZE },
      { "report",             required_argument, 0, 'r' },
      { "check-files",        no_argument,       0, LONGOPT_MKIMAGECHECK },
      { "no-check-files",     no_argument,       0, LONGOPT_NOMKIMAGECHECK },
      { "servers-section",    no_argument,       0, LONGOPT_ADDSERVERS },
      { "no-servers-section", no_argument,       0, LONGOPT_NOADDSERVERS },
      { "image-section",      no_argument,       0, LONGOPT_ADDIMAGE },
      { "no-image-section",   no_argument,       0, LONGOPT_NOADDIMAGE },
      { "label",              required_argument, 0, LONGOPT_LABEL },
      { "uri",                required_argument, 0, LONGOPT_URI },
      { 0, 0, 0, 0 }
    };

    int c = getopt_long(argc, argv, "0123456789hHvT:i:j:t:c:fr:",
                        longopts, 0);
    if (c == -1) break;

    switch (c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      optZipQuality = c - '0'; break;
    case 'h': case 'H': optHelp = c; break;
    case 'v': optVersion = true; break;
    case 'T': fileNames.addFilesFrom(
                strcmp(optarg, "-") == 0 ? "" : optarg); break;
    case 'i': imageFile = optarg; break;
    case 'j': jigdoFile = optarg; break;
    case 't': templFile = optarg; break;
    case 'c': cacheFile = optarg; break;
    case LONGOPT_NOCACHE: cacheFile.erase(); break;
    case LONGOPT_CACHEEXPIRY: optCacheExpiry = scanTimespan(optarg); break;
    case 'f': optForce = true; break;
    case LONGOPT_NOFORCE: optForce = false; break;
    case LONGOPT_MINSIZE:   miniBlockLength = scanMemSize(optarg); break;
    case LONGOPT_BLOCKSIZE:     blockLength = scanMemSize(optarg); break;
    case LONGOPT_MD5SIZE:    md5BlockLength = scanMemSize(optarg); break;
    case LONGOPT_BUFSIZE: readAmount = scanMemSize(optarg); break;
    case 'r':
      if (strcmp(optarg, "default") == 0) {
        optReporter = &reporterDefault;
      } else if (strcmp(optarg, "noprogress") == 0) {
        optReporter = &reporterNoprogress;
      } else if (strcmp(optarg, "quiet") == 0) {
        optReporter = &reporterQuiet;
      } else if (strcmp(optarg, "grep") == 0) {
        optReporter = &reporterGrep;
      } else {
        cerr << binaryName << ": Invalid argument to --report (allowed: "
          "default noprogress quiet grep)\n";
        error = true;
      }
      break;
    case LONGOPT_MKIMAGECHECK: optMkImageCheck = true; break;
    case LONGOPT_NOMKIMAGECHECK: optMkImageCheck = false; break;
    case LONGOPT_ADDSERVERS: optAddServers = true; break;
    case LONGOPT_NOADDSERVERS: optAddServers = false; break;
    case LONGOPT_ADDIMAGE: optAddImage = true; break;
    case LONGOPT_NOADDIMAGE: optAddImage = false; break;
    case LONGOPT_LABEL: optLabels.push_back(string(optarg)); break;
    case LONGOPT_URI: optUris.push_back(string(optarg)); break;
    case '?': error = true;
    case ':': break;
    default:
#     if DEBUG
      cerr << "getopt returned " << static_cast<int>(c) << endl;
#     endif
      break;
    }
  }

  if (error) exit_tryHelp();

  if (optHelp != '\0' || optVersion) {
    if (optVersion) cout << "jigdo-file version " JIGDO_VERSION << endl;
    if (optHelp != '\0') printUsage(optHelp == 'H');
    throw Cleanup(0);
  }
  //______________________________

  // Silently correct invalid miniBlockLength/blockLength/md5BlockLength args
  if (miniBlockLength < 4*1024) miniBlockLength = 4*1024;
  if (miniBlockLength > blockLength) blockLength = miniBlockLength;
  if (blockLength >= md5BlockLength) md5BlockLength = blockLength + 1;
  // Round to next k*64+55 for efficient MD5 calculation
  md5BlockLength = ((md5BlockLength + 63 - 55) & ~63U) + 55;

  Paranoid(miniBlockLength >= 4*1024
           && miniBlockLength <= blockLength
           && blockLength < md5BlockLength);
  //______________________________

  // Complain if name of command isn't there
  if (optind >= argc) {
    cerr << binaryName << ": Please specify a command\n";
    exit_tryHelp();
  }

  // Find Command code corresponding to command string on command line :)
  Command result;
  {
    const char* command = argv[optind++];
    struct CodesEntry { char* name; Command code; };
    const CodesEntry codes[] = {
      { "make-template",  MAKE_TEMPLATE },
      { "mt",             MAKE_TEMPLATE },
      { "make-image",     MAKE_IMAGE },
      { "mi",             MAKE_IMAGE },
      { "print-missing",  PRINT_MISSING },
      { "pm",             PRINT_MISSING },
      { "verify",         VERIFY },
      { "ver",            VERIFY },
      { "scan",           SCAN },
      { "sc",             SCAN },
      { "list-template",  LIST_TEMPLATE },
      { "ls",             LIST_TEMPLATE }
    };

    const CodesEntry *c = codes;
    const CodesEntry *end = codes+sizeof(codes)/sizeof(CodesEntry);
    while (true) {
      if (strcmp(command, c->name) == 0) {
        result = c->code;
        break;
      }
      ++c;
      if (c == end) {
        cerr << binaryName << ": Invalid command `" << command << "'\n"
          "(Must be one of:";
        for (const CodesEntry *c = codes,
               *end = codes+sizeof(codes)/sizeof(CodesEntry); c != end; ++c)
          cerr << ' ' << c->name;
        cerr << ")\n";
        exit_tryHelp();
      }
    }
  }
  //____________________

  while (optind < argc) fileNames.addFile(argv[optind++]);

# if 0
  /* If no --files-from given and no files on command line, assume we
     are to read any list of filenames from stdin. */
  if (fileNames.empty()) fileNames.addFilesFrom("");
# endif
  //____________________

  // If --image, --jigdo or --template not given, create name from other args
  if (imageFile.empty())
    deduceName(imageFile, "", jigdoFile, templFile);
  if (jigdoFile.empty())
    deduceName(jigdoFile, EXTSEPS"jigdo", imageFile, templFile);
  if (templFile.empty())
    deduceName(templFile, EXTSEPS"template", imageFile, jigdoFile);
  //____________________

# if DEBUG
  cerr << "Image file:  " << imageFile << endl;
  cerr << "Jigdo:       " << jigdoFile << endl;
  cerr << "Template:    " << templFile << endl;
# endif

  return result;
}
//______________________________________________________________________

void outOfMemory() {
  cerr << binaryName << ": Out of memory - aborted." << endl;
  exit(3);
}
//______________________________________________________________________

/* Open named file or stdin if name is "-". Store pointer to stream
   obj in dest and return it (except when it points to stdin; in this
   case return null). */
istream* openForInput(istream*& dest, const string& name) throw(Cleanup) {
  if (name == "-") {
    /* EEEEK! There's no standard way to switch mode of cin to binary.
       (There might be an implementation-dependent way? close()ing and
       re-open()ing cin may have strange effects.) */
    dest = reinterpret_cast<istream*>(&cin);
    return 0;
  } else {
    dest = new ifstream(name.c_str(), ios::binary);
    if (!*dest) {
      cerr << binaryName << ": Error opening `" << name
           << "' for input (" << strerror(errno) << ')' << endl;
      throw Cleanup(3);
    }
    return dest;
  }
}

// bistream* openForInput(bistream*& dest, const string& name) throw(Cleanup) {
//   if (name == "-") {
//     dest = reinterpret_cast<bistream*>(&cin);
//     return 0;
//   } else {
//     dest = new bifstream(name.c_str(), ios::binary);
//     if (!*dest) {
//       cerr << binaryName << ": Error opening `" << name
//            << "' for input (" << strerror(errno) << ')' << endl;
//       throw Cleanup(3);
//     }
//     return dest;
//   }
// }

/* Ensure that an output file is not already present. Should use this
   for all files before openForOutput() */
int willOutputTo(const string& name, bool errorMessage = true) {
  if (optForce) return 0;
  struct stat fileInfo;
  int err = stat(name.c_str(), &fileInfo);
  if (err == -1 && errno == ENOENT) return 0;

  if (errorMessage) {
    cerr << binaryName << ": Output file `" << name
         << "' already exists - delete it or use --force" << endl;
  }
  return 1;
}

ostream* openForOutput(ostream*& dest, const string& name) throw(Cleanup) {
  if (name == "-") {
    dest = reinterpret_cast<ostream*>(&cout); // EEEEK!
    return 0;
  } else {
    dest = new ofstream(name.c_str(), ios::binary);
    if (!*dest) {
      cerr << binaryName << ": Error opening `" << name
           << "' for output (" << strerror(errno) << ')' << endl;
      throw Cleanup(3);
    }
    return dest;
  }
}

// bostream* openForOutput(bostream*& dest, const string& name) throw(Cleanup) {
//   if (name == "-") {
//     dest = reinterpret_cast<bostream*>(&cout); // EEEEK!
//     return 0;
//   } else {
//     dest = new bofstream(name.c_str(), ios::binary);
//     if (!*dest) {
//       cerr << binaryName << ": Error opening `" << name
//            << "' for output (" << strerror(errno) << ')' << endl;
//       throw Cleanup(3);
//     }
//     return dest;
//   }
// }
//______________________________________________________________________

/* Read contents of optLabels/optUris and call addLabel() for the
   supplied cache object to set up the label mapping.
   optLabels/optUris is cleared after use. */
int addLabels(JigdoCache& cache) {
  int result = 0;
  string path, label, uri;

  // Create a map from label name to URI
  map<string, string> uriMap;
  for (vector<string>::iterator i = optUris.begin(), e = optUris.end();
       i != e; ++i) {
    pair<string, string> entry;
    string::size_type firstEquals = i->find('=');
    if (firstEquals == string::npos) {
      cerr << binaryName << ": Invalid argument to --uri: `" << *i << "'\n";
      result = 1;
    }
    entry.first.assign(*i, 0U, firstEquals);
    entry.second.assign(*i, firstEquals + 1, string::npos);
    // Add mapping to uriMap
#   if DEBUG
    cerr << "URI mapping: `" << entry.first << "' => `" << entry.second
         << "'" << endl;
#   endif
    uriMap.insert(entry);
  }
  optUris.clear();

  // Go through list of --label arguments and add them to JigdoCache
  for (vector<string>::iterator i = optLabels.begin(), e = optLabels.end();
       i != e; ++i) {
    // Label name is everything before first '=' in argument to --label
    string::size_type firstEquals = i->find('=');
    if (firstEquals == string::npos) {
      cerr << binaryName << ": Invalid argument to --label: `" << *i
           << "'\n";
      result = 1;
    }
    label.assign(*i, 0U, firstEquals);
    path.assign(*i, firstEquals + 1, string::npos);
    map<string, string>::iterator m = uriMap.find(label); // Lookup
    if (m == uriMap.end()) {
      uri = "file:";
      uri += path;
      compat_swapFileUriChars(uri);
      if (uri[uri.length() - 1] != '/') uri += '/';
      ConfigFile::quote(uri);
    } else {
      uri = m->second;
    }
    cache.addLabel(path, label, uri);
  }
  optLabels.clear();
  return result;
}

/* As above, but add URIs to the beginning of the [Servers] section
   of a ConfigFile. rescan() is only necessary when changing section
   lines. */
void addUris(ConfigFile& config) {
  // Let ci point to the line before which the label mapping will be inserted
  ConfigFile::iterator ci = config.firstSection("Servers");
  if (ci == config.end()) {
    // No [Servers] section yet; append it at end
    config.insert(ci, string("[Servers]"));
    config.rescan();
  } else {
    ++ci;
  }

  for (vector<string>::iterator i = optUris.begin(), e = optUris.end();
       i != e; ++i) {
    // Just add it, no matter what it contains...
    config.insert(ci, *i);
  }
  optUris.clear();
  return;
}
//______________________________________________________________________

int makeTemplate() {
  if (imageFile.empty() || jigdoFile.empty() || templFile.empty()) {
    cerr << binaryName <<
      " make-template: Not all of --image, --jigdo, --template specified.\n"
      "(Attempt to deduce missing names failed.)\n";
    exit_tryHelp();
  }

  if (fileNames.empty()) {
    optReporter->info("Warning - no files specified. The template will "
                      "contain the complete image contents!");
  }

  if (imageFile != "-") {
    struct stat fileInfo;
    if (stat(imageFile.c_str(), &fileInfo) == 0)
      imageSize = fileInfo.st_size;
  }

  // Give >1 error messages if >1 output files not present, hence no "||"
  if (willOutputTo(jigdoFile) + willOutputTo(templFile) > 0)
    throw Cleanup(3);

  // Open files
  bistream* image;
  auto_ptr<bistream> imageDel(openForInput(image, imageFile));

  ostream* jigdo;
  auto_ptr<ostream> jigdoDel(openForOutput(jigdo, jigdoFile));

  bostream* templ;
  auto_ptr<bostream> templDel(openForOutput(templ, templFile));
  //____________________

  JigdoCache cache(cacheFile, optCacheExpiry, readAmount, *optReporter);
  cache.setParams(miniBlockLength, blockLength, md5BlockLength);
  if (addLabels(cache)) return 3;
  while (true) {
    try { cache.readFilenames(fileNames); } // Recurse through directories
    catch (RecurseError e) { optReporter->error(e.message); continue; }
    break;
  }
  auto_ptr<MkTemplate>
    op(new MkTemplate(&cache, image, jigdo, templ, *optReporter,
                      miniBlockLength, optZipQuality, readAmount,
                      optAddImage, optAddServers));
  size_t lastDirSep = imageFile.find_last_of(DIRSEP);
  if (lastDirSep == string::npos) lastDirSep = 0; else ++lastDirSep;
  string imageFileLeaf(imageFile, lastDirSep);
  if (op.get()->run(imageFileLeaf)) return 3;
  return 0;
}
//______________________________________________________________________

int makeImage() {
  if (imageFile.empty() || templFile.empty()) {
    cerr << binaryName <<
      " make-image: Not both --image and --template specified.\n"
      "(Attempt to deduce missing names failed.)\n";
    exit_tryHelp();
  }

  if (imageFile != "-" && willOutputTo(imageFile) > 0) return 3;
  JigdoCache cache(cacheFile, optCacheExpiry, readAmount, *optReporter);
  cache.setParams(miniBlockLength, blockLength, md5BlockLength);
  while (true) {
    try { cache.readFilenames(fileNames); } // Recurse through directories
    catch (RecurseError e) { optReporter->error(e.message); continue; }
    break;
  }

  string imageTmpFile;
  if (imageFile != "-") {
    imageTmpFile = imageFile;
    imageTmpFile += EXTSEPS"tmp";
  }

  bistream* templ;
  auto_ptr<bistream> templDel(openForInput(templ, templFile));

  try {
    return JigdoDesc::makeImage(&cache, imageFile, imageTmpFile, templFile,
      templ, optForce, *optReporter, readAmount, optMkImageCheck);
  } catch (Error e) {
    string err = binaryName; err += " make-image: "; err += e.message;
    optReporter->error(err);
    return 3;
  }
}
//______________________________________________________________________

int listTemplate() {
  if (templFile.empty()) {
    cerr << binaryName << " list-template: --template not specified.\n";
    exit_tryHelp();
  }
  if (templFile == "-") {
    cerr << binaryName << " list-template: Sorry, cannot read from stdin.\n";
    exit_tryHelp();
  }

  // Open file
  bistream* templ;
  auto_ptr<bistream> templDel(openForInput(templ, templFile));

  if (JigdoDesc::isTemplate(*templ) == false)
    optReporter->info("Warning: This does not seem to be a template file");

  JigdoDescVec contents;
  try {
    JigdoDesc::seekFromEnd(*templ);
    *templ >> contents;
    contents.list(cout);
    if (!*templ) {
      string err = binaryName;
      err += " list-template: "; err += strerror(errno);
      optReporter->error(err);
      return 3;
    }
  } catch (JigdoDescError e) {
    string err = binaryName; err += ": "; err += e.message;
    optReporter->error(err);
    return 3;
  }
  return 0;
}
//______________________________________________________________________

int verifyImage() {
  if (imageFile.empty() || templFile.empty()) {
    cerr << binaryName <<
      " verify: Not both --image and --template specified.\n"
      "(Attempt to deduce missing names failed.)\n";
    exit_tryHelp();
  }

  bistream* image;
  auto_ptr<bistream> imageDel(openForInput(image, imageFile));

  JigdoDescVec contents;
  JigdoDesc::ImageInfo* info;
  try {
    bistream* templ;
    auto_ptr<bistream> templDel(openForInput(templ, templFile));

    if (JigdoDesc::isTemplate(*templ) == false)
      optReporter->info("Warning: This does not seem to be a template file");

    JigdoDesc::seekFromEnd(*templ);
    *templ >> contents;
    if (!*templ) {
      string err = binaryName;
      err += " verify: "; err += strerror(errno);
      optReporter->error(err);
      return 3;
    }
    info = dynamic_cast<JigdoDesc::ImageInfo*>(contents.back());
    if (info == 0) {
      string err = binaryName;
      err += " verify: Invalid template data - corrupted file?";
      optReporter->error(err);
      return 3;
    }
  } catch (JigdoDescError e) {
    string err = binaryName; err += ": "; err += e.message;
    optReporter->error(err);
    return 3;
  }

  MD5Sum md; // MD5Sum of image
  md.updateFromStream(*image, info->size(), readAmount, *optReporter);
  md.finish();
  if (*image) {
    image->get();
    if (image->eof() && md == info->md5()) {
      optReporter->info("OK: Checksums match, image is good!");
      return 0;
    }
  }
  optReporter->error(
      "ERROR: Checksums do not match, image might be corrupted!");
  return 2;
}
//______________________________________________________________________

/* Look up a query (e.g. "MyServer:foo/path/bar") in the JigdoConfig
   mapping. Returns true if something was found, and prints out all
   resulting URIs. */
bool printMissing_lookup(JigdoConfig& jc, const string& query,
                         bool printAll) {
  JigdoConfig::Lookup l(jc, query);
  string uri;
  if (!l.next(uri)) return false;
  do {
    // Omit "file:" when printing
    if (uri[0] == 'f' && uri[1] == 'i' && uri[2] == 'l'
        && uri[3] == 'e' && uri[4] == ':') {
      string nativeFilename(uri, 5);
      compat_swapFileUriChars(nativeFilename);
      cout << nativeFilename << endl;
    } else {
      cout << uri << '\n';
    }
  } while (printAll && l.next(uri));
  return true;
}
//______________________________

int printMissing() {
  if (imageFile.empty() || jigdoFile.empty() || templFile.empty()) {
    cerr << binaryName <<
      " make-template: Not all of --image, --jigdo, --template specified.\n"
      "(Attempt to deduce missing names failed.)\n";
    exit_tryHelp();
  }

  bistream* templ;
  auto_ptr<bistream> templDel(openForInput(templ, templFile));

  string imageTmpFile;
  if (imageFile != "-") {
    imageTmpFile = imageFile;
    imageTmpFile += EXTSEPS"tmp";

    // If image file exists, assume that it is complete; print nothing
    struct stat fileInfo;
    int err = stat(imageFile.c_str(), &fileInfo);
    if (err == 0) return 0;
  }

  // Read .jigdo file
  istream* jigdo;
  auto_ptr<istream> jigdoDel(openForInput(jigdo, jigdoFile));
  auto_ptr<ConfigFile> cfDel(new ConfigFile());
  ConfigFile* cf = cfDel.get();
  *jigdo >> *cf;
  JigdoConfig jc(jigdoFile, cfDel.release(), *optReporter);
  // Add any mappings specified on command line
  if (!optUris.empty()) {
    addUris(jc.configFile());
    jc.rescan();
  }

  set<MD5> sums;
  try {
    JigdoDesc::listMissing(sums, imageTmpFile, templFile, templ,
                           *optReporter);
  } catch (Error e) {
    string err = binaryName; err += " print-missing: "; err += e.message;
    optReporter->error(err);
    return 3;
  }
  //____________________

# if 0
  // To just list the MD5 sums:
  for (set<MD5>::iterator i = sums.begin(), e = sums.end(); i != e; ++i)
    cout << *i << endl;
# endif

# if 0
  // To list all URIs for each missing file, separated by empty lines:
  for (set<MD5>::iterator i = sums.begin(), e = sums.end(); i != e; ++i) {
    stringstream mdStream;
    Base64ostream m(mdStream);
    m.write(i->sum, 16).flush();
    string s(mdStream.str());

    vector<string> words;
    size_t off;
    bool found = false;
    for (ConfigFile::Find f(cf, "Parts", s, &off);
         !f.finished(); off = f.next()) {
      // f.section() points to "[section]" line, or end() if 0th section
      // f.label()   points to "label=..." line, or end() if f.finished()
      // off is offset of part after "label=", or 0
      words.clear();
      ConfigFile::split(words, f.label(), off);
      // Ignore everything but the first word
      if (printMissing_lookup(jc, words[0], true)) found = true;
    }
    if (!found) {
      /* No mapping found in [Parts] (this shouldn't happen) - create
         fake "MD5sum:<md5sum>" label line */
      s.insert(0, "MD5Sum:");
      printMissing_lookup(jc, s, false);
    }
    cout << endl;
  }
# endif

  // To list just the first URI
  for (set<MD5>::iterator i = sums.begin(), e = sums.end(); i != e; ++i) {
    stringstream mdStream;
    Base64ostream m(mdStream);
    m.write(i->sum, 16).flush();
    string s(mdStream.str());

    vector<string> words;
    size_t off;
    bool found = false;
    for (ConfigFile::Find f(cf, "Parts", s, &off);
         !f.finished(); off = f.next()) {
      // f.section() points to "[section]" line, or end() if 0th section
      // f.label()   points to "label=..." line, or end() if f.finished()
      // off is offset of part after "label=", or 0
      words.clear();
      ConfigFile::split(words, f.label(), off);
      // Ignore everything but the first word
      if (printMissing_lookup(jc, words[0], false)) { found = true; break; }
    }
    if (!found) {
      /* No mapping found in [Parts] (this shouldn't happen) - create
         fake "MD5sum:<md5sum>" label line */
      s.insert(0, "MD5Sum:");
      printMissing_lookup(jc, s, false);
    }
  }

  return 0;
}
//______________________________________________________________________

// Enter all file arguments into the cache
int scanFiles() {
  if (cacheFile.empty()) {
    cerr << binaryName << " scan: Please specify a --cache file.\n";
    exit_tryHelp();
  }

  JigdoCache cache(cacheFile, optCacheExpiry, readAmount, *optReporter);
  cache.setParams(miniBlockLength, blockLength, md5BlockLength);
  if (addLabels(cache)) return 3;
  while (true) {
    try { cache.readFilenames(fileNames); } // Recurse through directories
    catch (RecurseError e) { optReporter->error(e.message); continue; }
    break;
  }
  JigdoCache::iterator ci = cache.begin(), ce = cache.end();
  while (ci != ce) {
    ci->getMD5Sum(&cache); // Cause file to be read
    ++ci;
  }
  return 0;
  // Cache data is written out when the JigdoCache is destroyed
}
//______________________________________________________________________

} // local namespace

int main(int argc, char* argv[]) {
  try {
    set_new_handler(outOfMemory);
    Command command = cmdOptions(argc, argv);
    switch (command) {
    case MAKE_TEMPLATE:  returnValue = makeTemplate(); break;
    case MAKE_IMAGE:     returnValue = makeImage();    break;
    case PRINT_MISSING:  returnValue = printMissing(); break;
    case SCAN:           returnValue = scanFiles();    break;
    case VERIFY:         returnValue = verifyImage();  break;
    case LIST_TEMPLATE:  returnValue = listTemplate(); break;
    }
  }
  catch (bad_alloc) { outOfMemory(); }
  catch (Cleanup c) {
#   if DEBUG
    cerr << "[Cleanup " << c.returnValue << ']' << endl;
#   endif
    return c.returnValue;
  }
  catch (Error e) {
    string err = binaryName; err += ": "; err += e.message;
    optReporter->error(err);
    return 3;
  }
  catch (...) { // Uncaught exception - this should not happen(tm)
    string err = binaryName; err += ": Unknown error";
    optReporter->error(err);
    return 3;
  }
# if DEBUG
  cerr << "[exit(" << returnValue << ")]" << endl;
# endif
  return returnValue;
}
