/* $Id: mimestream.hh,v 1.8 2001/08/22 14:40:47 richard Exp $ -*- C++ -*-
  __   _
  |_) /|  Copyright (C) 2000 Richard Atterer
  | \/|  <richard@atterer.net>
   ` 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2. See
  the file COPYING for details.

  Convert binary data to/from ASCII using Base64 encoding

*/

#ifndef MIMESTREAM_HH
#define MIMESTREAM_HH

#ifndef INLINE
#  ifdef NOINLINE
#    define INLINE
#    else
#    define INLINE inline
#  endif
#endif

#include <iostream>
namespace std { }
using namespace std;

#include <config.h>
#include <debug.hh>
//______________________________________________________________________

/** Convert binary data to Base64 while outputting. Note that this
    does *not* implement the RFC2045 requirement that lines of text be
    no longer than 76 characters each. Furthermore, by default the
    data is not terminated with any '='. */
class Base64ostream { // No need to derive from ostream, just imitate it!
public: ///
  Base64ostream() : stream(0) { } ///
  ~Base64ostream() { flush(); } ///
  explicit Base64ostream(ostream& s) : bits(0), stream(&s) { } ///
  bool is_open() const { return stream != 0; } ///
  void open(ostream& s) { Assert(!is_open()); bits = 0; stream = &s; } ///
  void close(); ///

  /// Output operators
  Base64ostream& operator<<(char x) { return put(x); } ///
  Base64ostream& operator<<(signed char x) { return put(x); } ///
  Base64ostream& operator<<(unsigned char x) { return put(x); } ///
  /// Output the low 8 bits of an integer
  Base64ostream& operator<<(int x) { return put(x); }
  /// Output 32 bit integer in little-endian order
  Base64ostream& operator<<(uint32 x) { return put(x); }

  /// Output null-terminated string
  inline Base64ostream& operator<<(const char* x); ///
  inline Base64ostream& operator<<(const signed char* x); ///
  Base64ostream& operator<<(const unsigned char* x); ///
  inline Base64ostream& operator<<(const void* x);

  /// Get reference to underlying ostream
  ostream& getStream() { return *stream; }
  /** Implicit conversion to the underlying ostream always flushes the
      Base64ostream first */
  operator ostream&() { flush(); return getStream(); }

  /// Output 1 character
  inline Base64ostream& put(unsigned char x); ///
  inline Base64ostream& put(signed char x); ///
  /// Output the low 8 bits of an integer
  inline Base64ostream& put(int x); ///
  inline Base64ostream& put(char x);
  /// Output 32 bit integer in little-endian order
  Base64ostream& put(uint32 x);
  /// Output n characters
  inline Base64ostream& write(const char* x, size_t n); ///
  inline Base64ostream& write(const signed char* x, size_t n); ///
  Base64ostream& write(const unsigned char* x, size_t n); ///
  inline Base64ostream& write(const void* x, size_t n);

  /** Standard ostream operations are forwarded to actual stream. NB:
      This implementation is probably not 100% correct WRT the
      semantic definition of these functions. */
  bool good() const { return stream != 0 && stream->good(); } ///
  bool eof() const { return stream != 0 && stream->eof(); } ///
  bool fail() const { return stream != 0 && stream->fail(); } ///
  bool bad() const { return stream != 0 && stream->bad(); } ///
  operator void*() const {
    return reinterpret_cast<void*>(fail() ? -1 : 0); } ///
  bool operator!() const { return fail(); }
  /** flush() is not forwarded. Instead, it terminates the Base64
      string. NB: This is *not* a no-op. */
  Base64ostream& flush();
  /** Output the appropriate number of '=' characters (0, 1 or 2)
      given how many bytes were fed into the Base64ostream object. */
  Base64ostream& trailer(streamsize n);

private:
  static const byte code[];
  void putBits();
  int bits;
  uint32 data;
  ostream* stream;
};

// Convert Base64 to binary data while inputting
// class Base64istream {
//   ...
// };
//______________________________________________________________________

// Support functions to allow things like "b64Stream << endl;"

// Flush just the Base64ostream, not the underlying ostream
inline Base64ostream& flush(Base64ostream& s) {
  return s.flush();
}

/// Flush the Base64ostream, then forward endl to the ostream
inline Base64ostream& endl(Base64ostream& s) {
  s.flush();
  s.getStream() << endl;
  return s;
}

inline Base64ostream& operator<<(Base64ostream& s,
                                 Base64ostream& (*m)(Base64ostream&)) {
  return (*m)(s);
}
//______________________________________________________________________

Base64ostream& Base64ostream::operator<<(const char* x) {
  return (*this) << reinterpret_cast<const unsigned char*>(x);
}
Base64ostream& Base64ostream::operator<<(const signed char* x) {
  return (*this) << reinterpret_cast<const unsigned char*>(x);
}
Base64ostream& Base64ostream::operator<<(const void* x) {
  return (*this) << static_cast<const unsigned char*>(x);
}

Base64ostream& Base64ostream::put(unsigned char x) {
  data = (data << 8) | x;
  bits += 2;
  (*stream) << code[(data >> bits) & 63U];
  if (bits >= 8) {
    bits -= 6;
    (*stream) << code[(data >> bits) & 63U];
  }
  return *this;
}
Base64ostream& Base64ostream::put(signed char x) {
  return put(static_cast<unsigned char>(x));
}
Base64ostream& Base64ostream::put(char x) {
  return put(static_cast<unsigned char>(x));
}
Base64ostream& Base64ostream::put(int x) {
  return put(static_cast<unsigned char>(x));
}

Base64ostream& Base64ostream::write(const char* x, size_t n) {
  return write(reinterpret_cast<const unsigned char*>(x), n);
}
Base64ostream& Base64ostream::write(const signed char* x, size_t n) {
  return write(reinterpret_cast<const unsigned char*>(x), n);
}
Base64ostream& Base64ostream::write(const void* x, size_t n) {
  return write(static_cast<const unsigned char*>(x), n);
}

#endif
